import requests
import xmltodict
from repoze.lru import LRUCache, lru_cache

from utils import *

class BustimeError(Exception): pass
class NoPredictionsError(BustimeError): pass
class BustimeWarning(Exception): pass

class BustimeAPI(object):
    """
    A `requests` wrapper around the Port Authority's bustime API with
    some basic error handling and some useful conversions (like an
    option to convert all responses from JSON to XML.)
    
    This attempts to stay relatively close to the actual API's format with
    a few post-processing liberties taken to make things more Python friendly.
    
    Requires: `api_key`

    Optional: `locale` (defaults to en_US)
              `_format` (defaults to `json`, can be `xml`),
              `tmres` (time resolution, defaults to `s`)
              `cache` (cache non-dynamic information, defaults to False;
                       currently caches stops, routes)

    Implements: `bulletins`, `geopatterns`, `predictions`, `route_directions`, 
                `routes`, `stops`, `systemtime`, `vehicles 
    
    If you have an API key:
    >>> bus = PAACBustime(my_api_key)
    
    Official API documentation can be found at the Port Authority site:
    http://realtime.portauthority.org/bustime/apidoc/v1/main.jsp?section=documentation.jsp
    """
    
    __api_version__ = 'v1'
    
    ENDPOINTS = dict(
        SYSTIME = "http://realtime.portauthority.org/bustime/api/v1/gettime",
        VEHICLES = "http://realtime.portauthority.org/bustime/api/v1/getvehicles",
        ROUTES = "http://realtime.portauthority.org/bustime/api/v1/getroutes",
        R_DIRECTIONS = "http://realtime.portauthority.org/bustime/api/v1/getdirections",
        STOPS = "http://realtime.portauthority.org/bustime/api/v1/getstops",
        R_GEO = "http://realtime.portauthority.org/bustime/api/v1/getpatterns",
        PREDICTION = "http://realtime.portauthority.org/bustime/api/v1/getpredictions",
        BULLETINS = "http://realtime.portauthority.org/bustime/api/v1/getservicebulletins"
    )
    
    RESPONSE_TOKEN = "bustime-response"
    ERROR_TOKEN = "error"
    STRPTIME = "%Y%m%d %H:%M:%S"
    
    def __init__(self, apikey, locale="en_US", _format="json", tmres="s"):
        self.key = apikey
        self.format = _format
        self.args = dict(
            localestring = locale,
            tmres = tmres
        )
            
    def endpoint(self, endpt, argdict=None):
        """
        Construct API endpoint URLs using instance options in `self.args` 
        and local arguments passed to the function as a dictionary `argdict`.
        
        >>> api = PAACBustime("BOGUSAPIKEY")
        >>> api.endpoint('VEHICLES') 
        'http://realtime.portauthority.org/bustime/api/v1/getvehicles?key=BOGUSAPIKEY&tmres=s&localestring=en_US&format=json'
        >>> api.endpoint('PREDICTION', dict(stpid=4123, rt="61C"))
        'http://realtime.portauthority.org/bustime/api/v1/getpredictions?key=BOGUSAPIKEY&tmres=s&localestring=en_US&format=json&rt=61C&stpid=4123'
        """
        
        instanceargs = "{}&{}".format(queryjoin(key=self.key), queryjoin(self.args))
        if argdict:
            localargs = queryjoin(argdict)
            querystring = "{}&{}".format(instanceargs, localargs)
        else:    
            querystring = instanceargs
        return "{}?{}".format(self.ENDPOINTS[endpt], querystring)
        
    def response(self, url):
        """Validate and handle an API response."""
        
        def errorhandle(resp):            
            """Parse API error responses and raise appropriate exceptions."""
            if self.format == 'json':
                parsed = xmltodict.parse(resp)
                errors = parsed[self.RESPONSE_TOKEN][self.ERROR_TOKEN]
                # Create list of errors if more than one error response is given
                if type(errors) is list and len(errors) > 1:
                    messages = ", ".join([" ".join(["{}: {}".format(k,v) for k, v in e.items()]) for e in errors])
                else:
                    messages = " ".join(["{}: {}".format(k,v) for k, v in errors.items()])    
            elif self.format == 'xml':
                import xml.etree.ElementTree as ET
                errors = ET.fromstring(resp).findall(self.ERROR_TOKEN)
                messages = ", ".join(err.find('msg').text for err in errors)
            else:
                raise ValueError("Invalid API response format specified: {}." % self.format)        
            
            raise BustimeError("API returned: {}".format(messages))
                
        resp = requests.get(url).content
        if self.RESPONSE_TOKEN not in resp:
            raise BustimeError("The Bustime API returned an invalid response: {}".format(resp))
        elif self.ERROR_TOKEN in resp:
            return errorhandle(resp)
        else:
            if self.format == 'json':
                return xmltodict.parse(resp)[self.RESPONSE_TOKEN]
            elif self.format == 'xml':
                return resp
    
    def systemtime(self):
        """
        Get the API's official time (local, eastern).
        
        Arguments: none.
        
        Response:
            `tm`: "Date and time is represented in the following format: YYYYMMDD HH:MM:SS."
            
        http://realtime.portauthority.org/bustime/apidoc/v1/main.jsp?section=time.jsp    
        
        """
        return self.response(self.endpoint('SYSTIME'))
        
    def vehicles(self, vid=None, rt=None):
        """
        Get busses by route or by vehicle ID.
        
        Arguments: either
            `vid`: "Set of one or more vehicle IDs whose location should be returned." 
                   Maximum of 10 `vid`s, either in a comma-separated list or an iterable.

            `rt`: "Set of one or more route designators for which matching vehicles should be returned."
                  Maximum of 10 routes, either in a comma-separated list or an iterable.
                  
        Response:
            `vehicle`: (vehicle container) contains list of
                `vid`: bus #
                `tmstmp`: local date/time of vehicle update
                `lat`, `lon`: position
                `hdg`: vehicle heading (e.g., 0: north, 180: south)
                `pid`: pattern ID of current trip (see `self.geopatterns`)
                `pdist`: distance into trip
                `rt`: route (e.g, 88)
                `des`: bus destinations (e.g., "Penn to Bakery Square")
                `dly` (optional): True if bus is delayed
                `spd`: speed in mph
                `zone`: current zone (usually `None` here)
                `tablockid`, `tatripid`: unsure, seems internal?
                
        http://realtime.portauthority.org/bustime/apidoc/v1/main.jsp?section=vehicles.jsp
        """
        
        if vid and rt:
            raise ValueError("The `vid` and `route` parameters cannot be specified simultaneously.")
        if not (vid or rt):
            raise ValueError("You must specify either the `vid` or `rt` parameter.")

        # Turn list into comma separated string
        if listlike(rt): rt = ",".join( map(str, rt) )
        if listlike(vid): vid = ",".join( map(str, vid) )

        url = self.endpoint('VEHICLES', dict(vid=vid, rt=rt))        
        return self.response(url)
        
    @lru_cache(100)
    def routes(self):
        """
        Return a list of routes currently tracked by the API.
        
        Arguments: none

        Response:
            `route`: (route container) contains list of
                `rt`: route designator (e.g, P1, 88)
                `rtnm`: route name (e.g. EAST BUSWAY-ALL STOPS, PENN)
                `rtclr`: color of route used in map display (e.g. #9900ff),
                         usually unimportant                 
                         
        http://realtime.portauthority.org/bustime/apidoc/v1/main.jsp?section=routes.jsp
        """
        return self.response(self.endpoint('ROUTES'))
        
    @lru_cache(100)
    def route_directions(self, rt):
        """
        Return a list of directions for a route.
        
        The directions seem to always be INBOUND and OUTBOUND for the busses
        currently, where INBOUND is towards downtown and OUTBOUND is away from
        downtown. (No idea if this is going to change.)
        
        Arguments:
            `rt`: route designator
        
        Response:
            list of `dir`: directions served (e.g., INBOUND, OUTBOUND)    
            
        http://realtime.portauthority.org/bustime/apidoc/v1/main.jsp?section=routeDirections.jsp    
        """
        url = self.endpoint('R_DIRECTIONS', dict(rt=rt))
        return self.response(url)

    @lru_cache(1024)
    def stops(self, rt, direction):
        """
        Return a list of stops for a particular route.
        
        Arguments:
            `rt`: route designator
            `dir`: route direction (INBOUND, OUTBOUND)
        
        Response:
            `stop`: (stop container) contains list of 
                `stpid`: unique ID number for bus stop
                `stpnm`: stop name (what shows up on the display in the bus,
                         e.g., "Forbes and Murray")
                `lat`, `lng`: location of stop
                
        http://realtime.portauthority.org/bustime/apidoc/v1/main.jsp?section=stops.jsp        
        """
        url = self.endpoint('STOPS', dict(rt=rt, dir=direction))
        return self.response(url)
    
    def geopatterns(self, rt=None, pid=None):
        """
        Returns a set of geographic points that make up a particular routing
        ('pattern', in API terminology) for a bus route. A bus route can have
        more than one routing.
        
        Arguments: either
            `rt`: route designator
            or
            `pid`: comma-separated list or an iterable of pattern IDs (max 10)
            
        Response:
            `ptr`: (pattern container) contains list of
                `pid`: pattern ID
                `ln`: length of pattern in feet
                `rtdir`: route direction (see `self.route_directions`)
                
                `pt`: geo points, contains list of
                    `seq`: position of point in pattern
                    `typ`: 'S' = stop, 'W' = waypoint

                    -> if `typ` is a stop, will contain:
                    `stpid`, `stpnm`, `pdist` (see `self.stops`)
                    
                    `lat`, `lon`: position of point
                    
        http://realtime.portauthority.org/bustime/apidoc/v1/main.jsp?section=patterns.jsp   
        """
        if rt and pid:
            raise ValueError("The `rt` and `pid` parameters cannot be specified simultaneously.")
        if not (rt or pid):
            ValueError("You must specify either the `rt` or `pid` parameter.")

        if listlike(pid): pid = ",".join(pid)
        
        url = self.endpoint("R_GEO", dict(rt=rt, pid=pid))            
        
        return self._lru_geopatterns(url)    
    
    @lru_cache(1024)
    def _lru_geopatterns(self, url):
        # @lru_cache doesn't support kwargs so this is a bit of a workaround...
        return self.response(url)    
        
    def predictions(self, stpid="", rt="", vid="", maxpredictions=""):
        """
        Retrieve predictions for 1+ stops or 1+ vehicles.
        
        Arguments:
            `stpid`: unique ID number for bus stop (single or comma-seperated list or iterable)
            or
            `vid`: vehicle ID number (single or comma-seperated list or iterable)
            or
            `stpid` and `rt`
            
            `maxpredictions` (optional): limit number of predictions returned

        Response:
            `prd`: (prediction container) contains list of
                `tmstp`: when prediction was generated
                `typ`: prediction type ('A' = arrival, 'D' = departure)
                `stpid`: stop ID for prediction
                `stpnm`: stop name for prediction
                `vid`: vehicle ID for prediction
                `dstp`: vehicle distance to stop (feet)
                `rt`: bus route
                `des`: bus destination
                `prdtm`: ETA/ETD
                `dly`: True if bus delayed
                `tablockid`, `tatripid`, `zone`: internal, see `self.vehicles`
                
        http://realtime.portauthority.org/bustime/apidoc/v1/main.jsp?section=predictions.jsp    
        """
        
        if (stpid and vid) or (rt and vid):
            raise ValueError("These parameters cannot be specified simultaneously.")
        elif not (stpid or rt or vid):
            raise ValueError("You must specify a parameter.")   
        
        if listlike(stpid): stpid = ",".join(stpid)
        if listlike(rt): rt = ",".join(rt)
        if listlike(vid): vid = ",".join(vid)
                 
        if stpid or (rt and stpid) or vid:
            url = self.endpoint('PREDICTION', dict(rt=rt, stpid=stpid, vid=vid, top=maxpredictions))
            return self.response(url)
            
    def bulletins(self, rt="", rtdir="", stpid=""):
        """
        Return list of service alerts ('bulletins') for a route or stop.
        
        Arguments:
            `rt`: route designator
            or
            `stpid`: bus stop number
            or (`rt` and `rtdir`) or (`rt` and `rtdir` and `stpid`)
        
        Response:
             `sb`: (bulletin container) contains list of
                 `nm`: bulletin name/ID
                 `sbj`: bulletin subject
                 `dtl`: full text and/or
                 `brf`: short text
                 `prty`: priority (high, medium, low) 
                 `srvc`: (routes bulletin applies to) contains list of 
                     `rt`: route designator
                     `rtdir`: route direction
                     `stpid`: bus stop ID number
                     `stpnm`: bus stop name
        
        http://realtime.portauthority.org/bustime/apidoc/v1/main.jsp?section=serviceBulletins.jsp
        """
        
        if not (rt or stpid) or (rtdir and not (rt or stpid)):
            raise ValueError("You must specify a parameter.")   

        if listlike(stpid): stpid = ",".join(stpid)
        if listlike(rt): rt = ",".join(rt)
        
        url = self.endpoint('BULLETINS', dict(rt=rt, rtdir=rtdir, stpid=stpid))    
        return self.response(url)
        
        


from datetime import datetime, timedelta
from repoze.lru import lru_cache
from collections import namedtuple 
from pytz import timezone

from utils import listlike
from interface import BustimeError, BustimeWarning

class Bus(object):
    """Represents an individual vehicle on a route with a location."""
    
    @classmethod
    @lru_cache(5, timeout=10) # Cache for slightly less than the location refresh time
    def get(_class, api, vid):
        """
        Return a Bus object for a certain vehicle ID `vid` using API
        instance `api`.
        """
        busses = api.vehicles(vid=vid)['vehicle']   
        return _class.fromapi(api, api.vehicles(vid=vid)['vehicle'])
        
    @classmethod
    def fromapi(_class, api, apiresponse):
        """
        Return a Bus object from an API response dict.
        """
        bus = apiresponse
        return _class(
            api = api,
            vid = bus['vid'],
            timeupdated = datetime.strptime(bus['tmstmp'], api.STRPTIME),
            lat = float(bus['lat']),
            lng = float(bus['lon']),
            heading = bus['hdg'],
            pid = bus['pid'],
            intotrip = bus['pdist'],
            route = bus['rt'],
            destination = bus['des'],
            speed = bus['spd'],
            delay = bus.get('dly') or False
        )             

    def __init__(self, api, vid, timeupdated, lat, lng, heading, pid, intotrip, route, destination, speed, delay=False):
        self.api = api
        self.vid = vid
        self.timeupdated = timezone("US/Eastern").localize(timeupdated)
        self.location = (lat, lng)
        self.heading = int(heading)
        self.patternid = pid
        self.dist_in_trip = intotrip
        self.route = route
        self.destination = destination
        self.speed = speed
        self.delayed = delay
        
    def __str__(self):
        return "<Bus #{} on {} {}> - at {} as of {}".format(self.vid, self.route, self.destination, self.location, self.timeupdated)
        
    def __repr__(self):
        return self.__str__()
    
    def update(self):
        """Update this bus by creating a new one and transplanting dictionaries."""
        vehicle = self.api.vehicles(vid=self.vid)['vehicle']
        newbus = self.fromapi(self.api, vehicle)
        self.__dict__ = newbus.__dict__
        del newbus
    
    @property
    def pattern(self):
        return self.api.geopatterns(pid=self.patternid)
    
    @property
    def predictions(self):
        """Generator that yields prediction objects from an API response."""
        for prediction in self.api.predictions(vid=self.vid)['prd']:
            pobj = Prediction.fromapi(self.api, prediction)
            pobj._busobj = self
            yield pobj
                
    @property
    @lru_cache(1, timeout=10)
    def next_stop(self):
        """Return the next stop for this bus."""
        p = self.api.predictions(vid=self.vid)['prd']
        pobj = Prediction.fromapi(self.api, p[0])
        pobj._busobj = self
        return pobj
        
class OfflineBus(Bus): 
    """Sometimes, busses can be still present in the tracking system but not be
    reporting live data back, which causes the API to fail. This object is a
    representation for that case so the prediction function does not throw an
    exception."""       
    
    def __init__(self, vid):
        self.vid = vid
        
    def __str__(self):
        return "<Bus #{}: NO DATA [Location Currently Offline]>".format(self.vid)
        
        
class Route(object):
    """Represents a certain bus route (e.g. P1). Also contains a list of all the valid routes."""
    all_routes = {} # store list of all routes currently
    
    @classmethod
    def update_list(_class, rtdicts):
        allroutes = {}
        for rtdict in rtdicts:
            rtobject = Route.fromapi(api, rtdict)
            allroutes[str(rtobject.number)] = rtobject
        return allroutes
    
    @classmethod
    def get(_class, api, rt):
        """
        Return a Route object for route `rt` using API instance `api`.
        """
             
        if not _class.all_routes:
            _class.all_routes = self.update_list(api.routes()['route'])

        return _class.all_routes[str(rt)]
        
    @classmethod
    def fromapi(_class, api, apiresponse):
        return _class(api, apiresponse['rt'], apiresponse['rtnm'])
        
    def __init__(self, api, number, name):
        self.api = api
        self.number = number
        self.name = name
        self.stops = {}
        
    def __str__(self):
        return "{} {}".format(self.number, self.name)
    
    def __repr__(self):
        classname = self.__class__.__name__
        return "{}({}, {})".format(classname, self.name, self.number)
        
    def __hash__(self):
        return hash(str(self))

    @property
    @lru_cache(2, timeout=60*60*2)
    def bulletins(self):
        apiresponse = self.api.bulletins(rt=self.number)
        if apiresponse:
            for b in apiresponse['sb']:
                yield Bulletin.fromapi(b)
    
    @property
    def patterns(self):
        return self.api.geopatterns(rt=self.number)
    
    @property
    def busses(self):
        apiresp = self.api.vehicles(rt=self.number)['vehicle']
        if type(apiresp) is list:
            for busdict in apiresp:
                busobj = Bus.fromapi(self.api, busdict)
                busobj.route = self
                yield busobj
        else:
                busobj = Bus.fromapi(self.api, busdict)
                busobj.route = self
                yield busobj
        
    @property
    def directions(self):
        if not hasattr(self, "_directions"):
            self._directions = self.api.route_directions(self.number)['dir']
        return self._directions    
        
    @property    
    def inbound_stops(self):
        try:
            return self.stops['inbound']
        except:    
            inboundstops = self.api.stops(self.number, "INBOUND")['stop']
            self.stops['inbound'] = [StopWithLocation.fromapi(self.api, stop) for stop in inboundstops]
            return self.stops['inbound']
        
    @property    
    def outbound_stops(self):
        try:
            return self.stops['outbound']
        except:    
            outboundstops = self.api.stops(self.number, "OUTBOUND")['stop']
            self.stops['outbound'] = [StopWithLocation.fromapi(self.api, stop) for stop in outboundstops]
            return self.stops['outbound']
            
    def find_stop(self, query, direction=""):
        """
        Search the list of stops, optionally in a direction (inbound or outbound),
        for the term passed to the function. Case insensitive, searches both the
        stop name and ID. Yields a generator.
        
        Defaults to both directions.
        """
        _directions = ["inbound", "outbound", ""]
        direction = direction.lower()
        if direction == "inbound":
            stops = self.inbound_stops
        elif direction == "outbound":
            stops = self.outbound_stops
        else:
            stops = self.inbound_stops + self.outbound_stops
        
        found = []
        for stop in stops:
            q = str(query).lower()
            if q in stop.name.lower() or q in str(stop.id).lower():
                found.append(stop)
        return found
    
class Stop(object):
    """Represents a single stop."""
    
    @classmethod
    def get(_class, api, stpid):
        """
        Returns a Stop object for stop # `stpid` using API instance `api`.
        The API doesn't support looking up information for an individual stop,
        so all stops generated using Stop.get only have stop ID # info attached.
        
        Getting a stop from a Route is the recommended method of finding a specific
        stop (using `find_stop`/`inbound_stop`/`outbound_stop` functions).
        
        >>> Stop.get(1605)
        Stop(1605, "(Unnamed)")
        """
        return _class(api, stpid, "(Unnamed)")
            
    def __init__(self, api, _id, name):
        self.api = api
        self.id = _id
        self.name = name        
        
    def __repr__(self):
        classname = self.__class__.__name__
        return "{}({}, {})".format(classname, self.id, self.name)
        
    def __hash__(self):
        return hash(str(self))
    
    def predictions(self, route=''):
        """
        Yields predicted bus ETAs for this stop.  You can specify a 
        route identifier for ETAs specific to one route, or leave `route`
        blank (done by default) to get information on all arriving busses.
        """   
        apiresponse = self.api.predictions(stpid=self.id, rt=route)['prd']
        if type(apiresponse) is list:        
            for prediction in apiresponse:
                try:
                    pobj = Prediction.fromapi(self.api, prediction)
                    pobj._stopobj = self
                    yield pobj
                except:
                    continue    
        else:
            pobj = Prediction.fromapi(self.api, apiresponse)
            pobj._stopobj = self
            yield pobj
                
    @property
    @lru_cache(2, timeout=60*60*2)
    def bulletins(self):
        apiresponse = self.api.bulletins(stpid=self.id)
        if apiresponse:
            for b in apiresponse['sb']:
                yield Bulletin.fromapi(b)

class StopWithLocation(Stop):      
    """Represents a Stop with an added location parameter."""
    
    def get(self):
        raise NotImplementedError
        
    @classmethod
    def fromapi(_class, api, apiresponse):
        location = (apiresponse['lat'], apiresponse['lon'])
        stpid = apiresponse['stpid']
        # There might not be names occasionally
        name = apiresponse.get('stpnm') 
        return _class(api, stpid, name, location)
        
    def __init__(self, api, _id, name, location):
        super(StopWithLocation, self).__init__(api, _id, name)
        self.location = location
    
    def __str__(self):
        name = self.name or "(Unnamed)" 
        return "<Stop #{} {} at {}>".format(self.id, name, self.location)
    
    def __repr__(self):
        classname = self.__class__.__name__
        return "{}({}, {}, {})".format(classname, self.id, self.name, self.location)            
        
class Prediction(object):
    """Represents an ETA or ETD prediction for a certain Bus and/or Stop."""
    
    pstop = namedtuple("predicted_stop", ['id', 'name', 'feet_to'])

    @classmethod
    def fromapi(_class, api, apiresponse):
        generated_time = datetime.strptime(apiresponse['tmstmp'], api.STRPTIME)
        arrival = True if apiresponse['typ'] == 'A' else False
        bus = apiresponse['vid']
        stop = _class.pstop(apiresponse['stpid'], apiresponse['stpnm'], int(apiresponse['dstp']))
        route = apiresponse['rt']
        direction = apiresponse['rtdir']
        destination = apiresponse['des']
        et = datetime.strptime(apiresponse['prdtm'], api.STRPTIME)
        delayed = bool(apiresponse.get('dly'))
        
        return _class(api, et, arrival, delayed, generated_time, stop, route, destination, bus, direction)
        
    def __init__(self, api, eta, is_arrival, delayed, generated, stop, route, destination, bus, direction):
        self.api = api
        self.eta = timezone("US/Eastern").localize(eta) # Datetime ETA
        self.is_arrival = is_arrival # Is
        self.delayed = delayed
        self.generated = timezone("US/Eastern").localize(generated)
        self._stop = stop
        self.route, self.destination = route, destination
        self.direction = direction
        self._vid = bus

    def __str__(self):
        phrase = "ETA" if self.is_arrival else "ETD"
        return "<Prediction> {}: {} Bus: {} Stop: {}".format(phrase, self.eta, self.bus, self.stop, self.freshness)
    
    def __repr__(self):
        return str(self)
        
    @property
    def bus(self):
        if not hasattr(self, "_busobj"):
            try:
                self._busobj = Bus.fromapi(self.api, self.api.vehicles(vid=self._vid)['vehicle'])
            except BustimeError:
                self._busobj = OfflineBus(self._vid)  
        return self._busobj    
        
    @property
    def stop(self):
        if not hasattr(self, "_stopobj"):
            self._stopobj = Stop(self.api, self._stop.id, self._stop.name)
        return self._stopobj
        
    @property
    def dist_to_stop(self):
        return self._stop.feet_to    
        
    @property
    def freshness(self):
        now = datetime.now(timezone("US/Eastern"))
        change = divmod((now - self.generated).total_seconds(), 60)
        return timedelta(minutes=change[0], seconds=change[1])
        
class Bulletin(object):    
    """
    A service bulletin, usually representing a detour or other type of
    route change.
    """
    affected_service = namedtuple('affected_service', ['type', 'id', 'name'])
    @classmethod
    def get(_class, api, rt=None, rtdir=None, stpid=None):        
        if not (rt or stpid) or (rtdir and not (rt or stpid)):
            raise ValueError("You must specify a parameter.")   

        if listlike(stpid): stpid = ",".join(stpid)
        if listlike(rt): rt = ",".join(rt)
        
        bulletins = api.bulletins(rt=rt, rtdir=rtdir, stpid=stpid)
        
        if bulletins:
            bulletins = bulletins['sb']
            if type(bulletins) is list:
                return [_class.fromapi(b) for b in bulletins]
            else:
                return _class.fromapi(bulletins)        
    
    @classmethod 
    def fromapi(_class, apiresponse):
        """Create a bulletin object from an API response (dict), containing `sbj`, etc."""
        
        # Extract details from dict
        _id = "n/a" or apiresponse.get("nm")
        subject = apiresponse.get("sbj")
        text = apiresponse.get('dtl') + "\n" + apiresponse.get('brf')
        priority = "n/a" or apiresponse.get('prty')
        for_stops, for_routes = [], []
        svc = apiresponse.get('srvc')
        
        # Create list of affected routes/stops, if there are any
        if svc:
            has_stop = 'stpid' in svc or 'stpnm' in svc
            has_rt = 'rt' in svc or 'rtdir' in svc
            
            if has_stop: 
                aff = _class.affected_service('stop', svc.get('stpid'), svc.get('stpnm'))
                for_stops.append(aff)
            if has_rt:
                aff = _class.affected_service('route', svc.get('rt'), svc.get('rtdir'))
                for_routes.append(aff)

        return _class(_id, subject, text, priority, for_stops, for_routes)
        
    def __init__(self, _id, subject, text, priority, for_stops=None, for_routes=None):
        self.id = _id
        self.subject = subject
        self.body = text
        self.priority = priority
        self._stops = for_stops or []
        self._routes = for_routes or []
        
    def __str__(self):
        formatted = "Bulletin #{}\n\nSubject: {}\nPriority: {}\n\n{}\n\nValid for stops: {}\nValid for routes: {}"
        return formatted.format(self.id, self.subject, self.priority, self.body, self._stops, self._routes)
        
    @property
    def valid_for(self):
        return dict(
            stops=self._stops,
            routes=self._routes
        )
        


"""Some utility functions and other stuff."""

class CachedFunction(object):
    """Generate an LRU cached function given a cache object."""
    
    def __init__(self, cache):
        self.cache = cache

    def __call__(self, f):
        cache = self.cache
        marker = _MARKER
        def lru_cached(*arg):
            val = cache.get(arg, marker)
            if val is marker:
                val = f(*arg)
                cache.put(arg, val)
            return val
        lru_cached.__module__ = f.__module__
        lru_cached.__name__ = f.__name__
        lru_cached.__doc__ = f.__doc__
        return lru_cached
        
def queryjoin(argdict=dict(), **kwargs):
    """Turn a dictionary into a querystring for a URL.
    
    >>> args = dict(a=1, b=2, c="foo")
    >>> queryjoin(args)
    "a=1&b=2&c=foo"
    """
    if kwargs: argdict.update(kwargs)
    
    if issubclass(type(argdict), dict):                    
        args = ["{}={}".format(k, v) for k, v in argdict.items() if v != None]
    return "&".join(args)
    
def listlike(obj):
    """Is an object iterable like a list (and not a string)?"""
    
    return hasattr(obj, "__iter__") \
    and not issubclass(type(obj), str)\
    and not issubclass(type(obj), unicode)

def patterntogeojson(pattern):
    from geojson import Point, Feature, FeatureCollection
    
    """
    Turns an an API response of a pattern into a GeoJSON FeatureCollection.
    Takes a dict that contains at least `pid`, `ln`, `rtdir`, and `pt`.
    
    >>> api_response = {'ln': '123.45', 'pid': '1', 'pt': [], 'rtdir': 'OUTBOUND'}
    >>> pt1 = {'lat': '40.449', 'lon': '-79.983', 'seq': '1', 'typ': 'W'}
    >>> pt2 = {'stpid': '1', 'seq': '2', 'stpnm': '3142 Test Ave FS', 'lon': '-79.984', 'pdist': '42.4', 'lat': '40.450', 'typ': 'S'}
    >>> api_response['pt'] = [pt1, pt2]
    >>> patterntogeojson(api_response) # doctest: +ELLIPSIS
    {"features": [{"geometry": {"coordinates": ... "name": "3142 Test Ave FS", "type": "stop"}, "type": "Feature"}], "type": "FeatureCollection"}
    """ 
    # Base properties for the pattern
    properties = dict(
        pid = pattern['pid'],
        length = pattern['ln'],
        direction = pattern['rtdir']
    )
    
    pattern_features = []
    
    for point in pattern['pt']:
        # Base properties for each point
        pointprops = dict(
            i = int(point.get('seq')),
            type = "waypoint" if point.get('typ') == 'W' else "stop"
        )
        location = (float(point.get('lon')), float(point.get('lat')))
        # If it's a stop, add stop information
        if pointprops['type'] == 'stop':
            stop = dict(
                id = int(point.get('stpid')),
                name = point.get('stpnm'),
                dist_into_route = float(point.get('pdist'))
            )
            pointprops.update(stop)
        # Create a Feature to encapsulate the point and properties    
        pattern_features.append( Feature(geometry=Point(location), properties=pointprops ) )
    
    return FeatureCollection(pattern_features)    


from setuptools import setup, find_packages
setup(
    name="pgh-bustime",
    version='0.2.0',
    author='Nathaniel Fruchter',
    author_email='fruchter@cmu.edu',
    packages=['pghbustime'],
    url='http://github.com/nhfruchter/pgh-bustime',
    license='LICENSE',
    description='Python wrapper for the Port Authority of Allegheny County realtime bus information API.',
    install_requires=['xmltodict', 'requests', 'repoze.lru', 'pytz']
)
